#!/usr/bin/env python
#-------------------------------------------------------------------------------
# Name:         VirtualKeyboard
# Purpose:      Find optimal fingering for piano scores
# URL:          https://github.com/marcomusy/pianoplayer
# Author:       Marco Musy
#-------------------------------------------------------------------------------
from __future__ import division, print_function

try:
    import time
    from vtkplotter import Plotter, printc
    from vtkplotter.shapes import Ellipsoid, Box, Cylinder, Text
    from vtkplotter import Assembly
except:
    print("VirtualKeyboard: cannot find vtkplotter package. Not installed?")
    print('Try:\n(sudo) pip install --upgrade vtkplotter')
    quit()

from pianoplayer import __version__
from pianoplayer.utils import fpress, frelease, kpress, krelease, nameof
from pianoplayer.wavegenerator import playSound
import pianoplayer.utils as utils


###########################################################
class VirtualKeyboard:

    def __init__(self, songname=''):
                         
        self.KB = dict()
        self.vp = None
        self.rightHand = None
        self.leftHand  = None
        self.vpRH = None
        self.vpLH = None
        self.playsounds = True
        self.verbose = True
        self.songname = songname
        self.t0 = 0 # keep track of how many seconds to play
        self.dt = 0.1
        self.speedfactor = 1
        self.engagedfingersR = [False]*6 # element 0 is dummy
        self.engagedfingersL = [False]*6 
        self.engagedkeysR    = []
        self.engagedkeysL    = []

        self.build_keyboard() 

    #######################################################
    def makeHandActor(self, f=1):
        a1, a2, a3, c = (10*f,0,0), (0,7*f,0), (0,0,3*f), (.7,0.3,0.3)
        palm = Ellipsoid(pos=(0,-3,0), axis1=a1, axis2=a2, axis3=a3, alpha=0.6, c=c)
        wrist= Box(pos=(0,-9,0), length=6*f, width=5, height=2, alpha=0.4, c=c)
        arm  = Assembly([palm,wrist])
        self.vp.actors.append(arm) # add actor to internal list
        f1 = self.vp.add(Cylinder((-2, 1.5,0), axis=(0,1,0), height=5, r=.8*f, c=c))
        f2 = self.vp.add(Cylinder((-1, 3  ,0), axis=(0,1,0), height=6, r=.7*f, c=c))
        f3 = self.vp.add(Cylinder(( 0, 4  ,0), axis=(0,1,0), height=6.2, r=.75*f, c=c))
        f4 = self.vp.add(Cylinder(( 1, 3.5,0), axis=(0,1,0), height=6.1, r=.7*f, c=c))
        f5 = self.vp.add(Cylinder(( 2, 2  ,0), axis=(0,1,0), height=5, r=.6*f, c=c))
        return [arm, f1,f2,f3,f4,f5]

    def build_RH(self, hand):    
        if self.verbose: print('Building Right Hand..')
        self.rightHand = hand
        f = utils.handSizeFactor(hand.size)
        self.vpRH = self.makeHandActor(f)
        for limb in self.vpRH: # initial x positions are superseded later
            limb.x( limb.x()* 2.5 )
            limb.addPos([16.5*5+1, -7.5, 3] ) # vtkplotter < 8.7.1 was addpos()

    def build_LH(self, hand): #########################
        if self.verbose: print('Building Left Hand..')
        self.leftHand = hand
        f = utils.handSizeFactor(hand.size)
        self.vpLH = self.makeHandActor(f)
        for limb in self.vpLH: 
            limb.x( limb.x()* 2.5 ) 
            limb.addPos([16.5*3+1, -7.5, 3] ) # vtkplotter < 8.7.1 was addpos()
               

    #######################################################
    def build_keyboard(self):
        
        if self.verbose: print('Building Keyboard..')
        nts = ("C","D","E","F","G","A","B")
        tol = 0.12
        keybsize = 16.5 # in cm, span of one octave
        wb = keybsize/7
        nr_octaves = 7
        span = nr_octaves*wb*7
    
        self.vp = Plotter(title='PianoPlayer '+__version__, axes=0, size=(700,1400), bg='lb', verbose=0)

        #wooden top and base
        self.vp.add(Box(pos=(span/2+keybsize, 6,  1), length=span+1, height=3, width= 5).texture('wood5')) #top
        self.vp.add(Box(pos=(span/2+keybsize, 0, -1), length=span+1, height=1, width=17).texture('wood5'))
        self.vp.add(Text('PianoPlayer '+__version__, pos=(18, 5.5, 2), depth=.7, c='w'))
        self.vp.add(Text('https://github.com/marcomusy/pianoplayer', pos=(105,4.8,2), depth=.7, c='w', s=.8))
        leggio = self.vp.add(Box(pos=(span/1.55,8,10), length=span/2, height=span/8, width=0.08, c=(1,1,0.9)))
        leggio.rotateX(-20)
        self.vp.add(Text('Playing\n\n'+self.songname, pos=[0,0,0], s=1.2, c='k').rotateX(70).pos([49,7,9]))

        for ioct in range(nr_octaves):
            for ik in range(7):              #white keys
                x  = ik * wb + (ioct+1)*keybsize +wb/2
                tb = self.vp.add(Box(pos=(x,-2,0), length=wb-tol, height=1, width=12, c='white'))
                self.KB.update({nts[ik]+str(ioct+1) : tb})
                if not nts[ik] in ("E","B"): #black keys
                    tn=self.vp.add(Box(pos=(x+wb/2,0,1), length=wb*.6, height=1, width=8, c='black'))
                    self.KB.update({nts[ik]+"#"+str(ioct+1) : tn})
        self.vp.show(interactive=0)
        self.vp.camera.Azimuth(4)
        self.vp.camera.Elevation(-30)


    #####################################################################
    def play(self):
        printc('Press [0-9] to proceed by one note or for more seconds', c=1)
        printc('Press Esc to exit.', c=1)
        self.vp.keyPressFunction = self.runTime    # enable observer

        if self.rightHand:
            self.engagedkeysR    = [False]*len(self.rightHand.noteseq)
            self.engagedfingersR = [False]*6  # element 0 is dummy
        if self.leftHand:           
            self.engagedkeysL    = [False]*len(self.leftHand.noteseq)
            self.engagedfingersL = [False]*6        

        t=0.0
        while True:
            if self.rightHand: self._moveHand( 1, t)
            if self.leftHand:  self._moveHand(-1, t)
            if t > 1000: break                                         
            t += self.dt                      # absolute time flows
        
        if self.verbose: printc('End of note sequence reached.')
        self.vp.keyPressFunction = None       # disable observer

    ###################################################################
    def _moveHand(self, side, t):############# runs inside play() loop
        if side == 1: 
            c1,c2 = 'tomato', 'orange'
            engagedkeys    = self.engagedkeysR
            engagedfingers = self.engagedfingersR
            H              = self.rightHand
            vpH            = self.vpRH
        else:               
            c1,c2 = 'purple', 'mediumpurple'
            engagedkeys    = self.engagedkeysL
            engagedfingers = self.engagedfingersL
            H              = self.leftHand
            vpH            = self.vpLH

        for i, n in enumerate(H.noteseq):##################### 
            start, stop, f = n.time, n.time+n.duration, n.fingering
            if isinstance(f, str): continue
            if f and stop <= t <= stop+self.dt and engagedkeys[i]: #release key
                engagedkeys[i]    = False
                engagedfingers[f] = False
                name = nameof(n)
                krelease(self.KB[name])
                frelease(vpH[f])
                self.vp.interactor.Render()

        for i, n in enumerate(H.noteseq):##################### 
            start, stop, f = n.time, n.time+n.duration, n.fingering   
            if isinstance(f, str):
                print('Warning: cannot understand lyrics:',f, 'skip note',i)
                continue
            if f and start <= t < stop and not engagedkeys[i] and not engagedfingers[f]: #press key
                if i >= len(H.fingerseq): return                    
                engagedkeys[i]    = True
                engagedfingers[f] = True
                name = nameof(n)
                
                if t> self.t0 + self.vp.clock: 
                    self.t0 = t
                    self.vp.show(zoom=2, interactive=True)

                for g in [1,2,3,4,5]: 
                    vpH[g].x( side * H.fingerseq[i][g] ) 
                vpH[0].x(vpH[3].x()) # index 0 is arm, put it where middle finger is
                
                fpress(vpH[f],  c1)
                kpress(self.KB[name], c2)
                self.vp.show(zoom=2, interactive=False)

                if self.verbose:
                    msg = 'meas.'+str(n.measure)+' t='+str(round(t,2))
                    if side==1: printc(msg,'\t\t\t\tRH.finger', f, 'hit', name, c='b')
                    else:       printc(msg,      '\tLH.finger', f, 'hit', name, c='m')

                if self.playsounds: 
                    playSound(n, self.speedfactor)
                else: 
                    time.sleep(n.duration*self.speedfactor)


    ##########################################
    def runTime(self, key):
        secs = [str(i) for i in range(10)]
        if key not in secs: return
        printc('Will execute score for '+key+' seconds')
        self.vp.interactive = False
        self.vp.clock = int(key)
        self.vp.interactor.ExitCallback()


############################ test
if __name__ == "__main__":

    vk = VirtualKeyboard('Chopin Valse in A minor')
    vk.build_LH(None)

    fpress(vk.vpLH[1], 'b')
    kpress(vk.KB['E3'], 'db')
    vk.vp.show(zoom=2, interactive=1)

    frelease(vk.vpLH[1])
    krelease(vk.KB['E3'])
    vk.vp.show(zoom=2, interactive=1)








#
# 
#-------------------------------------------------------------------------------
# Name:         PianoPlayer
# Purpose:      Find optimal fingering for piano scores
# Author:       Marco Musy
#-------------------------------------------------------------------------------
from __future__ import division, print_function
from music21.articulations import Fingering
import pianoplayer.utils as utils


#####################################################
class Hand:
    def __init__(self, side="right", size='M'):
        
        self.LR        = side
        self.frest     = [None,  -7.0,-2.8, 0.0, 2.8, 5.8] # first is dummy, (cm)
        self.weights   = [None,   1.1, 1.0, 1.1, 0.9, 0.8] # finger rel strength
        self.bfactor   = [None,   0.3, 1.0, 1.1, 0.8, 0.7] # hit of black key bias
        self.noteseq   = []
        self.fingerseq = []
        self.depth     = 9
        self.autodepth = True  
        self.verbose   = True
        self.lyrics    = False  # show fingering numbers as lyrics in musescore
        self.size      = size
        self.handstretch = False

        self.hf = utils.handSizeFactor(size)
        for i in (1,2,3,4,5): 
            if self.frest[i]: self.frest[i] *= self.hf
        print('Your hand span set to size-'+size, 'which is', 21*self.hf, 'cm')
        print('(max relaxed distance between thumb and pinkie)')
        self.cfps = list(self.frest) # hold current finger positions


    #####################################################
    def set_fingers_positions(self, fings, notes, i): 

        fi = fings[i]
        ni = notes[i]
        fac = 1
        
        if self.handstretch:
            # allow whole hand stretching from -20% to +40%, 
            # based on the distribution of next 2 notes
            tfac, minfac, maxfac = 1, 0.8, 1.4
            nn = min(len(fings), len(notes), 2)
            if not ni.isChord or (self.LR=='left' and len(ni.chord21.pitches)==2):
                xs = [n.x for n in notes[i:i+nn]]
                if xs:
                    handspread = max(xs) - min(xs)
                    if handspread: 
                        tfac = handspread/(self.frest[5]-self.frest[1])*self.hf/0.8
                        if   tfac < minfac: fac = minfac 
                        elif tfac > maxfac: fac = maxfac
                        else: fac = abs(tfac)

        ifx = self.frest[fi]
        if ifx is not None:
            for j in (1,2,3,4,5): 
                self.cfps[j] = (self.frest[j]-ifx) *fac + ni.x
         
        
    #####################################################
    def ave_velocity(self, fingering, notes):
        ###calculate v for playing for notes in a given fingering combination
        
        self.set_fingers_positions(fingering, notes, 0)

        vmean = 0.
        for i in range(1, self.depth):
            na = notes[i-1]
            nb = notes[i]
            fb = fingering[i]
                        
            dx = nb.x - self.cfps[fb]     # space travelled by finger fb
            dt = nb.time - na.time +0.001 # available time +smoothing term
            v  = abs(dx/dt)               # velocity
            
            if nb.isBlack: 
                bfac = self.bfactor[fb]
            else: 
                bfac = 1
            v /= self.weights[fb] * bfac  # penalty (increase speed)
            vmean += v
            
            #update all fingers positions
            self.set_fingers_positions(fingering, notes, i)

        return vmean / (self.depth-1)
        
        
    #####################################################
    def optimize_seq(self, nseq, istart):
        '''Generate meaningful fingerings for a note sequence of size depth'''
    
        if self.autodepth:
            for i in (4,5,6,7,8):
                self.depth = i+1
                if nseq[i].time - nseq[0].time > 3: 
                    break #depth limit in secs
        depth = self.depth
    
        fingers = (1,2,3,4,5)
        n1, n2, n3, n4, n5 = nseq[0:5] 
        n6, n7, n8, n9 = [None]*4
        if depth>5: n6 = nseq[5]
        if depth>6: n7 = nseq[6]
        if depth>7: n8 = nseq[7]
        if depth>8: n9 = nseq[8]
        if istart == 0: u1 = fingers
        else: u1 = [istart]

        def skip(fa,fb, na,nb): ### two-consecutive-notes movement skipping rules ###
            # fa is fingering for note na
            xba = nb.x - na.x  # physical distance, cm

            if fa==fb and xba and na.duration<4 and not nb.isChord: 
                return True # play different notes w/ same finger, skip
        
            if fa>1 : # if a is not thumb
                if fb>1  and (fb-fa)*xba<0: return True # non-thumb fingers are crossings, skip
                if fb==1 and nb.isBlack and xba>0: return True # crossing thumb goes to black, skip
                
            if na.isChord and na.time-nb.time<0.0051: # na and nb are in the same chord
                axba = abs(xba)*self.hf/0.8 # max normalizd distance in cm btw 2 consecutive fingers
                if fa==fb  and axba: return True # play different chord notes w/ same finger, skip
                if axba> 6 and (fa==2 and fb==3 or fa==3 and fb==2): return True
                if axba>12 and (fa==1 and fb==2 or fa==2 and fb==1): return True
                if axba>14 and (fa==1 and fb==3 or fa==3 and fb==1): return True
                if axba>16 and (fa==1 and fb==4 or fa==4 and fb==1): return True
                if  xba> 0 and fa>1  and fb==1: return True # no thumb cross inside chord, skip
                if  xba< 0 and fa==1 and fb>1 : return True
            return False   ##########################################################

        out = ([0]*depth, -1)
        minvel = 1.e+10
        for f1 in u1:
            for f2 in fingers:
                if skip(f1,f2, n1,n2): continue
                for f3 in fingers:
                    if skip(f2,f3, n2,n3): continue
                    for f4 in fingers:
                        if skip(f3,f4, n3,n4): continue
                        for f5 in fingers:
                            if skip(f4,f5, n4,n5): continue
                            if depth<6: u=[False]
                            else: u=fingers
                            for f6 in u:
                                if f6 and skip(f5,f6, n5,n6): continue
                                if depth<7: u=[False]
                                else: u=fingers
                                for f7 in u:
                                    if f7 and skip(f6,f7, n6,n7): continue
                                    if depth<8: u=[False]
                                    else: u=fingers
                                    for f8 in u:
                                        if f8 and skip(f7,f8, n7,n8): continue
                                        if depth<9: u=[False]
                                        else: u=fingers
                                        for f9 in u:
                                            if f9 and skip(f8,f9, n8,n9): continue
                                            c = [f1,f2,f3,f4,f5,f6,f7,f8,f9]
                                            v = self.ave_velocity(c, nseq)
                                            if v < minvel: 
                                                out = (c, v)
                                                minvel  = v
        return out

    
    ###########################################################################################
    def generate(self, start_measure=0, nmeasures=1000):

        if start_measure == 1: start_measure=0 # avoid confusion with python numbering
        if self.LR == "left":
            for anote in self.noteseq:
                anote.x = -anote.x             # play left as a right on a mirrored keyboard

        start_finger, out, vel = 0, [0]*9, 0
        N = len(self.noteseq)
        if self.depth < 2: self.depth = 2
        if self.depth > 9: self.depth = 9

        for i in range(N):##############

            an = self.noteseq[i]
            if an.measure:
                if an.measure < start_measure : continue
                if an.measure > start_measure + nmeasures : break

            if i > N-11:
                self.autodepth = False
                self.depth = 9

            best_finger = 0
            if i > N-10:
                if len(out)>1: best_finger = out.pop(1)
            else:
                ninenotes = self.noteseq[i : i+9]
                out, vel  = self.optimize_seq(ninenotes, start_finger)
                best_finger  = out[0]
                start_finger = out[1]

            an.fingering = best_finger
            self.set_fingers_positions(out, ninenotes, 0)
            self.fingerseq.append(list(self.cfps))

            if best_finger>0:
                fng = Fingering(best_finger)
                if an.isChord:
                    if self.lyrics:
                         if len(an.chord21.pitches) <= 3:
                             # dont show fingering in the lyrics line for >3 note-chords
                             nl = len(an.chord21.pitches) - an.chordnr
                             an.chord21.addLyric(best_finger, nl)
                    else:
                        an.chord21.articulations.append(fng)
                else:
                    if self.lyrics: 
                        an.note21.addLyric(best_finger)
                    else:           
                        an.note21.articulations.append(fng)


            #-----------------------------
            if self.verbose:
                print("meas."+str(an.measure), end=' ')
                print("  finger:"+str(best_finger) + " on " + an.name+str(an.octave), end=' ')
                if i < N-10:
                    print("\tv=" + str(round(vel,1)), end='')
                    if self.autodepth:
                        print("\t"+"   "+str(out[0:self.depth]) + " d:" + str(self.depth))
                    else:
                        print("\t"+("   "*(i%self.depth))+str(out[0:self.depth]))
                else:
                    print()
            else:
                if i and not i%100:
                    print('scanned',i,'notes in', an.measure+1, 'measures for', self.LR ,'hand..')




